#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/types.h>
#include <sys/wait.h>

#define pi 3.14159265358979323846264338327950288419716939937510582097494459230781640628620899862803482534211706798214808651328

typedef struct {
	long mtype;
	char mtext[70];
} MSGBUF;

int status, pid1, pid2, msgid1, msgid2;

void Handler_C(int a)
{
	int rc;
	MSGBUF msg1, msg2;
	//signal(SIGINT, Handler_C);
	kill(pid1, SIGUSR1);
	kill(pid2, SIGUSR1);
	//4th parameter type = 0 -> read the first message in the queue
	//5th parameter flag = 0 -> msgrcv remains blocking until a message is found in queue
	rc = msgrcv(msgid1, &msg1, sizeof(msg1.mtext), 0, 0);
	if (rc == -1) {
		perror("first msgrcv");
		exit(1);
	}
	rc = msgrcv(msgid2, &msg2, sizeof(msg2.mtext), 0, 0);
	if (rc == -1) {
		perror("second msgrcv");
		exit(1);
	}
	printf("The difference of pi and Compute_pi1 is %.30f\n", atof(msg1.mtext)-pi);
	printf("The difference of pi and Compute_pi2 is %.30f\n", atof(msg2.mtext)-pi);
}

void Handler_Z(int a)
{
	int rc;
	MSGBUF msg1, msg2;
	//signal(SIGTSTP, Handler_Z);
	kill(pid1, SIGUSR1);
	kill(pid2, SIGUSR1);
	rc = msgrcv(msgid1, &msg1, sizeof(msg1.mtext), 0, 0);
        if (rc == -1) {
                perror("first msgrcv");
                exit(1);
        }
        rc = msgrcv(msgid2, &msg2, sizeof(msg2.mtext), 0, 0);
        if (rc == -1) {
                perror("second msgrcv");
                exit(1);
        }
 	if(atof(msg1.mtext)-pi < atof(msg2.mtext)-pi)
 		printf("First approach is faster\n");
 	else
 		printf("Second approach is faster\n");
		
 	kill(pid1, SIGKILL);
 	kill(pid2, SIGKILL);
}

int main ()
{
	int ret;
	
	 // Create message queues
    	msgid1 = msgget(10, IPC_CREAT | 0600);
    	msgid2 = msgget(20, IPC_CREAT | 0600);

    	if (msgid1 == -1) {
        	perror("msgget 1 in parent failed");
        	exit(EXIT_FAILURE);
    	}
	if (msgid2 == -1) {
		perror("msgget 2 in parent failed");
		exit(EXIT_FAILURE);
	}

	pid1=fork();
	if (pid1==-1)
	{
		printf("Error creating the first child \n");
		exit (1);
	}
	if (pid1==0)
	{
		ret=execl("./compute_pi1_2", "compute_pi1_2", NULL);
		if (ret==-1)
		{
			printf("Could not run process number one \n");
			exit (1);
		}
	}
	else
	{
		pid2=fork();
		if (pid2==-1)
		{
			printf("Error creating the second child \n");
			exit(1);
		}
		if (pid2==0)
		{
			ret=execl("./compute_pi2_2", "compute_pi2_2", NULL);
			if (ret==-1)
			{
				printf ("Could not run process number two \n ");
				exit(1);
			}
		}
		else
		{
			printf ("The program started succesfully\n");
			printf ("Press Ctrl^C to see the current values of pi and Ctrl^Z to terminate\n\n");
			signal(SIGINT, Handler_C);
			signal(SIGTSTP, Handler_Z);
			wait(&status);
			wait(&status);
			if (msgctl(msgid1, IPC_RMID, NULL) == -1) {
    				perror("msgctl 1 in parent failed");
    				exit(EXIT_FAILURE);
			} else {
    				printf("Message queue 1 deleted successfully.\n");
			}
			if (msgctl(msgid2, IPC_RMID, NULL) == -1) {
    				perror("msgctl 2 in parent failed");
    				exit(EXIT_FAILURE);
			} else {
    				printf("Message queue 2 deleted successfully.\n");
			}
		}
	}
}
